#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <usbdrv.h>

#define IDLE 0
#define UP 1 //pc3
#define DOWN 2 //pc4
#define RIGHT 3 //pc5
#define LEFT 4 //pc1

/* ------------------------------------------------------------------------- */
typedef struct 
{
	uint8_t modifier;  // แป้น Shift, Ctrl, Alt
	uint8_t keycode;   // แป้นทั่วไป
} keyboard_report_t;

keyboard_report_t reportBuffer; /* buffer for HID reports */
uchar idleRate = 10;            /* in 4 ms units */

/* ------------------------------------------------------------------------- */

PROGMEM const char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = 
{ /* USB report descriptor */
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};
/* We use a simplifed keyboard report descriptor which does not support the
 * boot protocol. We don't allow setting status LEDs and we only allow one
 * simultaneous key press (except modifiers). We can therefore use short
 * 2 byte input reports.
 * The report descriptor has been created with usb.org's "HID Descriptor Tool"
 * which can be downloaded from http://www.usb.org/developers/hidpage/.
 * Redundant entries (such as LOGICAL_MINIMUM and USAGE_PAGE) have been omitted
 * for the second INPUT item.
 */

/* Keyboard usage values, see usb.org's HID-usage-tables document, chapter
 * 10 Keyboard/Keypad Page for more codes.
 */
#define MOD_CONTROL_LEFT    (1<<0)
#define MOD_SHIFT_LEFT      (1<<1)
#define MOD_ALT_LEFT        (1<<2)
#define MOD_GUI_LEFT        (1<<3)
#define MOD_CONTROL_RIGHT   (1<<4)
#define MOD_SHIFT_RIGHT     (1<<5)
#define MOD_ALT_RIGHT       (1<<6)
#define MOD_GUI_RIGHT       (1<<7)

/*
 * เตรียมข้อความที่จะส่งไปยังโฮสท์
 * keycode ดึงมาจากเอกสาร Universial Serial Bus (USB) HID Usage Tables
 */
uint8_t output[] = {
	0, // idle
    0x52,  // keycode 'up'
    0x51,  // keycode 'down'
    0x4F,  // keycode 'right'
    0x50,  // keycode 'left'
};

/* ------------------------------------------------------------------------- */
/* ------------------------ interface to USB driver ------------------------ */
/* ------------------------------------------------------------------------- */

uchar usbFunctionSetup(uchar data[8])
{
    usbRequest_t *rq = (void *)data;

    if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
    {
        /* class request type */
        if (rq->bRequest == USBRQ_HID_GET_REPORT)
        {
            /* we only have one report type, so don't look at wValue */
            reportBuffer.modifier = 0;
            reportBuffer.keycode  = 0;
            usbMsgPtr = (void*)&reportBuffer;
            return sizeof(reportBuffer);
        }
        else if (rq->bRequest == USBRQ_HID_GET_IDLE)
        {
            usbMsgPtr = &idleRate;
            return 1;
        }
        else if (rq->bRequest == USBRQ_HID_SET_IDLE)
        {
            idleRate = rq->wValue.bytes[1];
        }
        
    }
	return 0;
}

void sendCode(uint8_t modifier, uint8_t keycode)
{
    reportBuffer.modifier = modifier;
    reportBuffer.keycode = keycode;
    usbSetInterrupt((uchar*)&reportBuffer, sizeof(reportBuffer));
}

/* ------------------------------------------------------------------------- */
/* --------------------------------- main ---------------------------------- */
/* ------------------------------------------------------------------------- */

int main(void)
{
    uchar STATE;

	DDRC &= ~((1<<PC1)|(1<<PC3)|(1<<PC4)|(1<<PC5)); //switch 1=left 5=right 3=up 4=down INPUT
	PORTC |= (1<<PC1)|(1<<PC3)|(1<<PC4)|(1<<PC5); //pullUp R all switch

    usbDeviceDisconnect();
    _delay_ms(300);
    usbDeviceConnect();

    usbInit();

    /* enable global interrupts */
    sei();

    /* main event loop */
    for(;;)
    {
        // press one button
        do { usbPoll(); } while (!usbInterruptIsReady());
        
		if((PINC&(1<<PC3))==0) STATE=UP;
        else if((PINC&(1<<PC4))==0) STATE=DOWN;
        else if((PINC&(1<<PC5))==0) STATE=RIGHT;
        else if((PINC&(1<<PC1))==0) STATE=LEFT;
        else STATE = IDLE;
		
		sendCode(0,output[STATE]);

        // ส่งคีย์โค้ด 0 เพื่อบอกโฮสท์ว่าคีย์ที่กดไปก่อนหน้านี้ถูกปล่อย
        do { usbPoll(); } while (!usbInterruptIsReady());
        sendCode(0,0);

    }

    return 0;
}
